
import primefac, functools
from hysop.tools.numerics import float_to_complex_dtype
from hysop.tools.numpywrappers import npw
from hysop.tools.decorators  import debug
from hysop.tools.warning import HysopWarning
from hysop.operator.base.curl import SpectralExternalForceOperatorBase
from hysop.backend.device.opencl.opencl_symbolic import OpenClSymbolic
from hysop.core.graph.graph import op_apply
from hysop.core.memory.memory_request import MemoryRequest
from hysop.backend.device.opencl.opencl_fft import OpenClFFT
from hysop.backend.device.codegen.base.variables import dtype_to_ctype
from hysop.symbolic import local_indices_symbols
from hysop.symbolic.relational import Assignment
from hysop.symbolic.complex import ComplexMul
from hysop.symbolic.misc import Select

class OpenClSpectralExternalForce(SpectralExternalForceOperatorBase, OpenClSymbolic):
    """
    Operator to compute the curl of a symbolic expression.
    """
    
    @debug
    def __init__(self, **kwds):
        """
        What is computed in 2D:
            FEXT_X = fft(Fext_X)
            tmp    = dFEXT/dy
            FEXT_Y = fft(Fext_Y)
            tmp   -= dFEXT/dx
            Fmin[0], Fmax[0], Finf[0] = min(tmp), max(tmp), max(abs(tmp))
            Wz  += dt*tmp
        What is computed in 3D:
            FEXT_X = fft(Fext_X)
            FEXT_Y = fft(Fext_Y)
            FEXT_Z = fft(Fext_Z)
            
            tmp = 
            Fmin[0], Fmax[0], Finf[0] = min(tmp), max(tmp), max(abs(tmp))
            Wx += dt*tmp

            tmp = 
            Fmin[1], Fmax[1], Finf[1] = min(tmp), max(tmp), max(abs(tmp))
            Wy += dt*tmp

            tmp = 
            Fmin[1], Fmax[1], Finf[1] = min(tmp), max(tmp), max(abs(tmp))
            Wz += dt*tmp
        """
        super(OpenClSpectralExternalForce, self).__init__(**kwds)

        assert (len(self.forward_transforms) % 2 == 0)
        N = len(self.forward_transforms)//2
        assert len(self.K)==2*N

        kernel_names = ()
        for (i,(Ft,(tg,Ki))) in enumerate(zip(self.forward_transforms, self.K)):
            Fhs = Ft.output_symbolic_array('F{}_hat'.format(i))
            
            kname = 'filter_curl_{}d_{}'.format(Fhs.dim, i)
            kernel_names += (kname,)
            
            is_complex = Ki.is_complex
            Ki = tg._indexed_wave_numbers[Ki]

            if is_complex:
                expr = ComplexMul(Ki, Fhs)
            else:
                expr = Ki*Fhs

            if (i<N):
                expr = Assignment(Fhs, +expr)
            else:
                expr = Assignment(Fhs, -expr)

            self.require_symbolic_kernel(kname, expr)

        self._kernel_names = kernel_names
    
    @debug
    def setup(self, work):
        super(OpenClSpectralExternalForce, self).setup(work)
        curl_filters = ()
        for kname in self._kernel_names:
            kernel, _ = self.symbolic_kernels[kname]
            kernel = functools.partial(kernel, queue=self.cl_env.default_queue)
            curl_filters += (kernel,)
        self.curl_filters = curl_filters
        self.exchange_ghosts = self.dFout.exchange_ghosts(build_launcher=True)
        assert len(self.forward_transforms)==len(self.backward_transforms)==len(curl_filters)

    @op_apply
    def apply(self, **kwds):
        """Solve the ExternalForce equation."""
        super(OpenClSpectralExternalForce, self).apply(**kwds)
        for (Ft,filter_curl,Bt) in zip(self.forward_transforms,
                                       self.curl_filters,
                                       self.backward_transforms):
            evt = Ft()
            evt = filter_curl()
            evt = Bt()
        if (self.exchange_ghosts is not None):
            evt = self.exchange_ghosts()
