
from abc import ABCMeta, abstractmethod
from hysop.fields
from hysop.symbolic.field import AppliedSymbolicField, SymbolicDiscreteField
from hysop.symbolic.parameter import SymbolicScalarParameter
from hysop.parameters.scalar_parameter import ScalarParameter
from hysop.operator.min_max import MinMaxFieldStatisticsBase

class ExternalForce(NamedObjectI):
    __metaclass__ = ABCMeta

    def __init__(self, name, dim, Fext, **kwds):
        super(ExternalForce, self).__init__(name=name, **kwds)

        check_instance(dim, int)
        self._dim  = dim
        self._Fext = Fext

        for f in self.input_fields:
            msg='Expected field dimension to be {} but got {} from field {}.'
            msg=msg.format(dim, field.dim, field.short_description())
            assert f.dim == dim, msg

    
    @property
    def name(self):
        return self._diffusion
    @property
    def pretty_name(self):
        return self._diffusion
    @property
    def Fext(self):
        return self._Fext
    @property
    def dim(self):
        return self._dim
    @property
    def diffusion(self):
        return self._diffusion

    @abstractmethod
    def input_fields(self):
        pass
    @abstractmethod
    def output_fields(self):
        pass
    @abstractmethod
    def input_params(self):
        pass
    @abstractmethod
    def output_params(self):
        pass


class SymbolicExternalForce(object):
    def __init__(self, name, Fext, diffusion=None, **kwds):
        """
        Specify an external force as a tuple of symbolic expressions.

        2D ExternalForce example:  
            1) Fext = -rho*g*e_y where rho is a field and g a constant
                Fext = (0, -rho.s*g)
            2) Fext = (Rs*S+C)*e_y
                Fext = (0, -Rs*S.s+C.s)
        """
        check_instance(Fext, tuple)
        dim = len(Fext)

        Fext = list(Fext)
        for i,e in enumerate(Fext):
            if isinstance(e, (type(None), int, long, float)):
                Fext[i] = 0 # curl(const) = 0
            msg='Expression "{}" contains a SymbolicField, did you forget to apply it ?'
            msg=msg.format(e)
            assert not e.atoms(SymbolicField), msg
        Fext = tuple(Fext)

        super(SymbolicExternalForce, self).__init__(name=name, dim=dim, Fext=Fext, **kwds)
        
        diffusion = first_not_None(diffusion, {})
        for (k,v) in diffusion.iteritems():
            assert k in self.input_fields
            assert isinstance(v, ScalarParameter)
        self._diffusion = diffusion

    def input_fields(self):
        return set(map(lambda f: f.field, self._extract_objects(AppliedSymbolicField)))
    def output_fields(self):
        return set(self._diffusion.keys())
    def input_params(self):
        p0 = set(map(lambda p: p.parameter, self._extract_objects(SymbolicScalarParameter)))
        p1 = set(self._diffusion.values())
        return p0.union(p1)
    def output_params(self):
        return set()

    def _extract_objects(self, obj_type):
        objs = ()
        for e in self.Fext:
            objs += e.atoms(obj_type)
        return objs


class SpectralExternalForce(ComputationalGraphNodeFrontend):
    """
    Generate an operator to compute the curl of a symbolic expression.
    """

    @classmethod
    def implementations(cls):
        from hysop.backend.device.opencl.operator.external_force import OpenClSpectralExternalForce
        __implementations = {
                Implementation.OPENCL: OpenClSpectralExternalForce,
        }
        return __implementations
    
    @classmethod
    def default_implementation(cls):
        return Implementation.OPENCL

    @debug
    def __init__(self, vorticity, Fext, dt, variables,
                    Fmin=None, Fmax=None, Finf=None,
                    implementation=None, base_kwds=None, **kwds):
        """
        Create an operator that computes the curl of a given input field Fext or a set of symbolic expressions.

        Only the following configurations are supported:
                     dim   nb_components  |   dim   nb_components
        vorticity:    2          1        |    3          3

        What is computed:
            tmp  = curl(Fext) by using a spectral backend
            Fmin = min(tmp)
            Fmax = max(tmp)
            Finf = max(abs(Fmin), abs(Fmax))
            W   += dt*tmp
        
        where Fext is computed from user given ExternalForce.

        Parameters
        ----------
        vorticity: hysop.field.continuous_field.Field
            Continuous field as input ScalarField or VectorField.
            All contained field have to live on the same domain.
        Fext: hysop.operator.external_force.ExternalForce
            Expression of the external force.
        F...: TensorParameter or boolean, optional
            TensorParameters should match the shape of tmp (see Notes).
            If set to True, the TensorParameter will be generated automatically.
            Autogenerated TensorParameters that are not required by the user are set to be quiet.
        pbasename: str, optional
            Parameters basename for created parameters.
            Defaults to 'curl_{}'.format(Fext.name).
        ppbasename: str, optional
            Parameters pretty basename for created parameters.
            Defaults to '|{} x {}|'.format(nabla, Fext.pretty_name).
        variables: dict
            dictionary of fields as keys and topologies as values.
        implementation: Implementation, optional, defaults to None
            target implementation, should be contained in available_implementations().
            If None, implementation will be set to default_implementation().
        kwds: dict, optional
            Extra parameters passed towards base class (MultiSpaceDerivatives).

        Notes
        -----
        If Fext.dim == 2, Fext
        """
        base_kwds = first_not_None(base_kwds, {})
        check_instance(vorticity, Field)
        check_instance(Fext, ExternalForce)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors)
        
        # Pregenerate parameters so that we can directly store them in self.
        default_pbasename  = 'curl_{}'.format(Fext.name)
        default_ppbasename = '{}_{}'.format(nabla, Fext.pretty_name)
        pbasename  = first_not_None(pbasename,  default_pbasename)
        ppbasename = first_not_None(ppbasename, default_ppbasename)
        parameters = MinMaxFieldStatisticsBase.build_parameters(field=field, 
                components=components, all_quiet=all_quiet,
                Fmin=Fmin, Fmax=Fmax, Finf=Finf, 
                pbasename=pbasename, ppbasename=ppbasename)

        (Fmin, Fmax, Finf) = tuple(parameters[k] for k in ('Fmin', 'Fmax', 'Finf'))

        super(Curl, self).__init__(Fin=Fin, Fout=Fout, variables=variables,
                candidate_input_tensors=(Fin,),
                candidate_output_tensors=(Fout,),
                implementation=implementation, 
                base_kwds=base_kwds, **kwds)

        self.Fmin = Fmin
        self.Fmax = Fmax
        self.Finf = Finf
